//
// BotMixer5000 - botman's multibot loader
//
// (http://planethalflife.com/botman/)
//
// dll.cpp
//

#include "extdll.h"
#include "enginecallback.h"
#include "util.h"
#include "cbase.h"

#include "bot_mixer.h"

DLL_FUNCTIONS other_gFunctionTable;

bot_t bots[32];
int current_dll_index;

extern GETENTITYAPI other_GetEntityAPI;
extern GETNEWDLLFUNCTIONS other_GetNewDLLFunctions;
extern enginefuncs_t g_engfuncs;
extern int debug_engine;
extern globalvars_t  *gpGlobals;
extern char *g_argv;

extern bot_dlls_t bot_dlls[MAX_DLLS];
extern int num_bot_dlls;

static FILE *fp;

int isFakeClientCommand = 0;
int fake_arg_count;

void BotFakeClientCommand(int bot_dll_index, edict_t *pEntity, char *cmd);


void FakeClientSet(int dll_index)
{
   current_dll_index = dll_index;

   for (int index = 0; index < 32; index++)
   {
      if (bots[index].pEdict)
      {
         if (bots[index].bot_dll_index == dll_index)
         {
            bots[index].pEdict->v.flags |= FL_FAKECLIENT;
            bots[index].pEdict->v.spawnflags |= FL_FAKECLIENT;
         }
         else
         {
            bots[index].pEdict->v.flags &= ~FL_FAKECLIENT;
            bots[index].pEdict->v.spawnflags &= ~FL_FAKECLIENT;
         }
      }
   }
}

void FakeClientRestore(void)
{
   if (current_dll_index < 0)  // is no dll currently active?
   {
      for (int index = 0; index < 32; index++)  // set all fake client's bits...
      {
         if (bots[index].pEdict)
         {
            bots[index].pEdict->v.flags |= FL_FAKECLIENT;
            bots[index].pEdict->v.spawnflags |= FL_FAKECLIENT;
         }
      }
   }
   else  // restore back to the current dll index settings...
   {
      for (int index = 0; index < 32; index++)
      {
         if (bots[index].pEdict)
         {
            if (bots[index].bot_dll_index == current_dll_index)
            {
               bots[index].pEdict->v.flags |= FL_FAKECLIENT;
               bots[index].pEdict->v.spawnflags |= FL_FAKECLIENT;
            }
            else
            {
               bots[index].pEdict->v.flags &= ~FL_FAKECLIENT;
               bots[index].pEdict->v.spawnflags &= ~FL_FAKECLIENT;
            }
         }
      }
   }
}


void GameDLLInit( void )
{
   int index;

   current_dll_index = -1;  // none in use currently

   for (index = 0; index < 32; index++)
   {
      bots[index].pEdict = NULL;
      bots[index].bot_dll_index = 0;
   }

   for (index = 0; index < num_bot_dlls; index++)
      (*bot_dlls[index].other_gFunctionTable.pfnGameInit)();

   (*other_gFunctionTable.pfnGameInit)();
}

int DispatchSpawn( edict_t *pent )
{
   int dll_index = current_dll_index;  // save current dll index

   for (int index = 0; index < num_bot_dlls; index++)
   {
      FakeClientSet(index);
      (*bot_dlls[index].other_gFunctionTable.pfnSpawn)(pent);
   }

   current_dll_index = dll_index;  // restore
   FakeClientRestore();

   return (*other_gFunctionTable.pfnSpawn)(pent);
}

void DispatchThink( edict_t *pent )
{
   int dll_index = current_dll_index;  // save current dll index

   for (int index = 0; index < num_bot_dlls; index++)
   {
      FakeClientSet(index);
      (*bot_dlls[index].other_gFunctionTable.pfnThink)(pent);
   }

   current_dll_index = dll_index;  // restore
   FakeClientRestore();

   (*other_gFunctionTable.pfnThink)(pent);
}

void DispatchUse( edict_t *pentUsed, edict_t *pentOther )
{
   int dll_index = current_dll_index;  // save current dll index

   for (int index = 0; index < num_bot_dlls; index++)
   {
      FakeClientSet(index);
      (*bot_dlls[index].other_gFunctionTable.pfnUse)(pentUsed, pentOther);
   }

   current_dll_index = dll_index;  // restore
   FakeClientRestore();

   (*other_gFunctionTable.pfnUse)(pentUsed, pentOther);
}

void DispatchTouch( edict_t *pentTouched, edict_t *pentOther )
{
   int dll_index = current_dll_index;  // save current dll index

   for (int index = 0; index < num_bot_dlls; index++)
   {
      FakeClientSet(index);
      (*bot_dlls[index].other_gFunctionTable.pfnTouch)(pentTouched, pentOther);
   }

   current_dll_index = dll_index;  // restore
   FakeClientRestore();

   (*other_gFunctionTable.pfnTouch)(pentTouched, pentOther);
}

void DispatchBlocked( edict_t *pentBlocked, edict_t *pentOther )
{
   int dll_index = current_dll_index;  // save current dll index

   for (int index = 0; index < num_bot_dlls; index++)
   {
      FakeClientSet(index);
      (*bot_dlls[index].other_gFunctionTable.pfnBlocked)(pentBlocked, pentOther);
   }

   current_dll_index = dll_index;  // restore
   FakeClientRestore();

   (*other_gFunctionTable.pfnBlocked)(pentBlocked, pentOther);
}

void DispatchKeyValue( edict_t *pentKeyvalue, KeyValueData *pkvd )
{
   int dll_index = current_dll_index;  // save current dll index

   for (int index = 0; index < num_bot_dlls; index++)
   {
      FakeClientSet(index);
      (*bot_dlls[index].other_gFunctionTable.pfnKeyValue)(pentKeyvalue, pkvd);
   }

   current_dll_index = dll_index;  // restore
   FakeClientRestore();

   (*other_gFunctionTable.pfnKeyValue)(pentKeyvalue, pkvd);
}

void DispatchSave( edict_t *pent, SAVERESTOREDATA *pSaveData )
{
   (*other_gFunctionTable.pfnSave)(pent, pSaveData);
}

int DispatchRestore( edict_t *pent, SAVERESTOREDATA *pSaveData, int globalEntity )
{
   return (*other_gFunctionTable.pfnRestore)(pent, pSaveData, globalEntity);
}

void DispatchObjectCollsionBox( edict_t *pent )
{
   int dll_index = current_dll_index;  // save current dll index

   for (int index = 0; index < num_bot_dlls; index++)
   {
      FakeClientSet(index);
      (*bot_dlls[index].other_gFunctionTable.pfnSetAbsBox)(pent);
   }

   current_dll_index = dll_index;  // restore
   FakeClientRestore();

   (*other_gFunctionTable.pfnSetAbsBox)(pent);
}

void SaveWriteFields( SAVERESTOREDATA *pSaveData, const char *pname, void *pBaseData, TYPEDESCRIPTION *pFields, int fieldCount )
{
   (*other_gFunctionTable.pfnSaveWriteFields)(pSaveData, pname, pBaseData, pFields, fieldCount);
}

void SaveReadFields( SAVERESTOREDATA *pSaveData, const char *pname, void *pBaseData, TYPEDESCRIPTION *pFields, int fieldCount )
{
   (*other_gFunctionTable.pfnSaveReadFields)(pSaveData, pname, pBaseData, pFields, fieldCount);
}

void SaveGlobalState( SAVERESTOREDATA *pSaveData )
{
   (*other_gFunctionTable.pfnSaveGlobalState)(pSaveData);
}

void RestoreGlobalState( SAVERESTOREDATA *pSaveData )
{
   (*other_gFunctionTable.pfnRestoreGlobalState)(pSaveData);
}

void ResetGlobalState( void )
{
   (*other_gFunctionTable.pfnResetGlobalState)();
}

BOOL ClientConnect( edict_t *pEntity, const char *pszName, const char *pszAddress, char szRejectReason[ 128 ]  )
{ 
   int dll_index = current_dll_index;  // save current dll index

   for (int index = 0; index < num_bot_dlls; index++)
   {
      FakeClientSet(index);
      (*bot_dlls[index].other_gFunctionTable.pfnClientConnect)(pEntity, pszName, pszAddress, szRejectReason);
   }

   current_dll_index = dll_index;  // restore
   FakeClientRestore();

   return (*other_gFunctionTable.pfnClientConnect)(pEntity, pszName, pszAddress, szRejectReason);
}

void ClientDisconnect( edict_t *pEntity )
{
   int index;
   int dll_index = current_dll_index;  // save current dll index

   for (index = 0; index < num_bot_dlls; index++)
   {
      FakeClientSet(index);
      (*bot_dlls[index].other_gFunctionTable.pfnClientDisconnect)(pEntity);
   }

   current_dll_index = dll_index;  // restore
   FakeClientRestore();

   // find this bot edict and remove it...
   for (index = 0; index < 32; index++)
   {
      if (bots[index].pEdict == pEntity)
      {
         bots[index].pEdict = NULL;
         break;
      }
   }

   (*other_gFunctionTable.pfnClientDisconnect)(pEntity);
}

void ClientKill( edict_t *pEntity )
{
   int dll_index = current_dll_index;  // save current dll index

   for (int index = 0; index < num_bot_dlls; index++)
   {
      FakeClientSet(index);
      (*bot_dlls[index].other_gFunctionTable.pfnClientKill)(pEntity);
   }

   current_dll_index = dll_index;  // restore
   FakeClientRestore();

   (*other_gFunctionTable.pfnClientKill)(pEntity);
}

void ClientPutInServer( edict_t *pEntity )
{
   int dll_index = current_dll_index;  // save current dll index

   for (int index = 0; index < num_bot_dlls; index++)
   {
      FakeClientSet(index);
      (*bot_dlls[index].other_gFunctionTable.pfnClientPutInServer)(pEntity);
   }

   current_dll_index = dll_index;  // restore
   FakeClientRestore();

   (*other_gFunctionTable.pfnClientPutInServer)(pEntity);
}

void ClientCommand( edict_t *pEntity )
{
   int dll_index, cnt, arg;
	const char *pcmd = CMD_ARGV(0);
   char cmd[256];

   for (dll_index = 0; dll_index < num_bot_dlls; dll_index++)
   {
      if (strcmp(bot_dlls[dll_index].cmd, pcmd) == 0)
      {
         cnt = (*g_engfuncs.pfnCmd_Argc)();

         cmd[0] = 0;

         // skip the bot MOD cmd and build the command line...
         for (arg = 1; arg < cnt; )
         {
            strcat(cmd, (*g_engfuncs.pfnCmd_Argv)(arg));

            arg++;

            if (arg < cnt)
               strcat(cmd, " ");
         }

         BotFakeClientCommand(dll_index, pEntity, cmd);

         return;  // don't call MOD ClientCommand for bot commands!!!
      }
   }

   (*other_gFunctionTable.pfnClientCommand)(pEntity);
}

void ClientUserInfoChanged( edict_t *pEntity, char *infobuffer )
{
   int dll_index = current_dll_index;  // save current dll index

   for (int index = 0; index < num_bot_dlls; index++)
   {
      FakeClientSet(index);
      (*bot_dlls[index].other_gFunctionTable.pfnClientUserInfoChanged)(pEntity, infobuffer);
   }

   current_dll_index = dll_index;  // restore
   FakeClientRestore();

   (*other_gFunctionTable.pfnClientUserInfoChanged)(pEntity, infobuffer);
}

void ServerActivate( edict_t *pEdictList, int edictCount, int clientMax )
{
   int dll_index = current_dll_index;  // save current dll index

   for (int index = 0; index < num_bot_dlls; index++)
   {
      FakeClientSet(index);
      (*bot_dlls[index].other_gFunctionTable.pfnServerActivate)(pEdictList, edictCount, clientMax);
   }

   current_dll_index = dll_index;  // restore
   FakeClientRestore();

   (*other_gFunctionTable.pfnServerActivate)(pEdictList, edictCount, clientMax);
}

void ServerDeactivate( void )
{
   int dll_index = current_dll_index;  // save current dll index

   for (int index = 0; index < num_bot_dlls; index++)
   {
      FakeClientSet(index);
      (*bot_dlls[index].other_gFunctionTable.pfnServerDeactivate)();
   }

   current_dll_index = dll_index;  // restore
   FakeClientRestore();

   (*other_gFunctionTable.pfnServerDeactivate)();
}

void PlayerPreThink( edict_t *pEntity )
{
   int dll_index = current_dll_index;  // save current dll index

   for (int index = 0; index < num_bot_dlls; index++)
   {
      FakeClientSet(index);
      (*bot_dlls[index].other_gFunctionTable.pfnPlayerPreThink)(pEntity);
   }

   current_dll_index = dll_index;  // restore
   FakeClientRestore();

   (*other_gFunctionTable.pfnPlayerPreThink)(pEntity);
}

void PlayerPostThink( edict_t *pEntity )
{
   int dll_index = current_dll_index;  // save current dll index

   for (int index = 0; index < num_bot_dlls; index++)
   {
      FakeClientSet(index);
      (*bot_dlls[index].other_gFunctionTable.pfnPlayerPostThink)(pEntity);
   }

   current_dll_index = dll_index;  // restore
   FakeClientRestore();

   (*other_gFunctionTable.pfnPlayerPostThink)(pEntity);
}

void StartFrame( void )
{
   int dll_index = current_dll_index;  // save current dll index

   for (int index = 0; index < num_bot_dlls; index++)
   {
      FakeClientSet(index);
      (*bot_dlls[index].other_gFunctionTable.pfnStartFrame)();
   }

   current_dll_index = dll_index;  // restore
   FakeClientRestore();

   (*other_gFunctionTable.pfnStartFrame)();
}

void ParmsNewLevel( void )
{
   (*other_gFunctionTable.pfnParmsNewLevel)();
}

void ParmsChangeLevel( void )
{
   (*other_gFunctionTable.pfnParmsChangeLevel)();
}

const char *GetGameDescription( void )
{
   return (*other_gFunctionTable.pfnGetGameDescription)();
}

void PlayerCustomization( edict_t *pEntity, customization_t *pCust )
{
   int dll_index = current_dll_index;  // save current dll index

   for (int index = 0; index < num_bot_dlls; index++)
   {
      FakeClientSet(index);
      (*bot_dlls[index].other_gFunctionTable.pfnPlayerCustomization)(pEntity, pCust);
   }

   current_dll_index = dll_index;  // restore
   FakeClientRestore();

   (*other_gFunctionTable.pfnPlayerCustomization)(pEntity, pCust);
}

void SpectatorConnect( edict_t *pEntity )
{
   (*other_gFunctionTable.pfnSpectatorConnect)(pEntity);
}

void SpectatorDisconnect( edict_t *pEntity )
{
   (*other_gFunctionTable.pfnSpectatorDisconnect)(pEntity);
}

void SpectatorThink( edict_t *pEntity )
{
   (*other_gFunctionTable.pfnSpectatorThink)(pEntity);
}

void Sys_Error( const char *error_string )
{
   (*other_gFunctionTable.pfnSys_Error)(error_string);
}

void PM_Move ( struct playermove_s *ppmove, int server )
{
   int dll_index = current_dll_index;  // save current dll index

   for (int index = 0; index < num_bot_dlls; index++)
   {
      FakeClientSet(index);
      (*bot_dlls[index].other_gFunctionTable.pfnPM_Move)(ppmove, server);
   }

   current_dll_index = dll_index;  // restore
   FakeClientRestore();

   (*other_gFunctionTable.pfnPM_Move)(ppmove, server);
}

void PM_Init ( struct playermove_s *ppmove )
{
   int dll_index = current_dll_index;  // save current dll index

   for (int index = 0; index < num_bot_dlls; index++)
   {
      FakeClientSet(index);
      (*bot_dlls[index].other_gFunctionTable.pfnPM_Init)(ppmove);
   }

   current_dll_index = dll_index;  // restore
   FakeClientRestore();

   (*other_gFunctionTable.pfnPM_Init)(ppmove);
}

char PM_FindTextureType( char *name )
{
   int dll_index = current_dll_index;  // save current dll index

   for (int index = 0; index < num_bot_dlls; index++)
   {
      FakeClientSet(index);
      (*bot_dlls[index].other_gFunctionTable.pfnPM_FindTextureType)(name);
   }

   current_dll_index = dll_index;  // restore
   FakeClientRestore();

   return (*other_gFunctionTable.pfnPM_FindTextureType)(name);
}

void SetupVisibility( edict_t *pViewEntity, edict_t *pClient, unsigned char **pvs, unsigned char **pas )
{
   int dll_index = current_dll_index;  // save current dll index

   for (int index = 0; index < num_bot_dlls; index++)
   {
      FakeClientSet(index);
      (*bot_dlls[index].other_gFunctionTable.pfnSetupVisibility)(pViewEntity, pClient, pvs, pas);
   }

   current_dll_index = dll_index;  // restore
   FakeClientRestore();

   (*other_gFunctionTable.pfnSetupVisibility)(pViewEntity, pClient, pvs, pas);
}

void UpdateClientData ( const struct edict_s *ent, int sendweapons, struct clientdata_s *cd )
{
   int dll_index = current_dll_index;  // save current dll index

   for (int index = 0; index < num_bot_dlls; index++)
   {
      FakeClientSet(index);
      (*bot_dlls[index].other_gFunctionTable.pfnUpdateClientData)(ent, sendweapons, cd);
   }

   current_dll_index = dll_index;  // restore
   FakeClientRestore();

   (*other_gFunctionTable.pfnUpdateClientData)(ent, sendweapons, cd);
}

int AddToFullPack( struct entity_state_s *state, int e, edict_t *ent, edict_t *host, int hostflags, int player, unsigned char *pSet )
{
   int dll_index = current_dll_index;  // save current dll index

   for (int index = 0; index < num_bot_dlls; index++)
   {
      FakeClientSet(index);
      (*bot_dlls[index].other_gFunctionTable.pfnAddToFullPack)(state, e, ent, host, hostflags, player, pSet);
   }

   current_dll_index = dll_index;  // restore
   FakeClientRestore();

   return (*other_gFunctionTable.pfnAddToFullPack)(state, e, ent, host, hostflags, player, pSet);
}

void CreateBaseline( int player, int eindex, struct entity_state_s *baseline, struct edict_s *entity, int playermodelindex, vec3_t player_mins, vec3_t player_maxs )
{
   (*other_gFunctionTable.pfnCreateBaseline)(player, eindex, baseline, entity, playermodelindex, player_mins, player_maxs);
}

void RegisterEncoders( void )
{
   (*other_gFunctionTable.pfnRegisterEncoders)();
}

int GetWeaponData( struct edict_s *player, struct weapon_data_s *info )
{
   int dll_index = current_dll_index;  // save current dll index

   for (int index = 0; index < num_bot_dlls; index++)
   {
      FakeClientSet(index);
      (*bot_dlls[index].other_gFunctionTable.pfnGetWeaponData)(player, info);
   }

   current_dll_index = dll_index;  // restore
   FakeClientRestore();

   return (*other_gFunctionTable.pfnGetWeaponData)(player, info);
}

void CmdStart( const edict_t *player, const struct usercmd_s *cmd, unsigned int random_seed )
{
   int dll_index = current_dll_index;  // save current dll index

   for (int index = 0; index < num_bot_dlls; index++)
   {
      FakeClientSet(index);
      (*bot_dlls[index].other_gFunctionTable.pfnCmdStart)(player, cmd, random_seed);
   }

   current_dll_index = dll_index;  // restore
   FakeClientRestore();

   (*other_gFunctionTable.pfnCmdStart)(player, cmd, random_seed);
}

void CmdEnd ( const edict_t *player )
{
   int dll_index = current_dll_index;  // save current dll index

   for (int index = 0; index < num_bot_dlls; index++)
   {
      FakeClientSet(index);
      (*bot_dlls[index].other_gFunctionTable.pfnCmdEnd)(player);
   }

   current_dll_index = dll_index;  // restore
   FakeClientRestore();

   (*other_gFunctionTable.pfnCmdEnd)(player);
}

int ConnectionlessPacket( const struct netadr_s *net_from, const char *args, char *response_buffer, int *response_buffer_size )
{
   return (*other_gFunctionTable.pfnConnectionlessPacket)(net_from, args, response_buffer, response_buffer_size);
}

int GetHullBounds( int hullnumber, float *mins, float *maxs )
{
   return (*other_gFunctionTable.pfnGetHullBounds)(hullnumber, mins, maxs);
}

void CreateInstancedBaselines( void )
{
   (*other_gFunctionTable.pfnCreateInstancedBaselines)();
}

int InconsistentFile( const edict_t *player, const char *filename, char *disconnect_message )
{
   return (*other_gFunctionTable.pfnInconsistentFile)(player, filename, disconnect_message);
}

int AllowLagCompensation( void )
{
   return (*other_gFunctionTable.pfnAllowLagCompensation)();
}


DLL_FUNCTIONS gFunctionTable =
{
   GameDLLInit,               //pfnGameInit
   DispatchSpawn,             //pfnSpawn
   DispatchThink,             //pfnThink
   DispatchUse,               //pfnUse
   DispatchTouch,             //pfnTouch
   DispatchBlocked,           //pfnBlocked
   DispatchKeyValue,          //pfnKeyValue
   DispatchSave,              //pfnSave
   DispatchRestore,           //pfnRestore
   DispatchObjectCollsionBox, //pfnAbsBox

   SaveWriteFields,           //pfnSaveWriteFields
   SaveReadFields,            //pfnSaveReadFields

   SaveGlobalState,           //pfnSaveGlobalState
   RestoreGlobalState,        //pfnRestoreGlobalState
   ResetGlobalState,          //pfnResetGlobalState

   ClientConnect,             //pfnClientConnect
   ClientDisconnect,          //pfnClientDisconnect
   ClientKill,                //pfnClientKill
   ClientPutInServer,         //pfnClientPutInServer
   ClientCommand,             //pfnClientCommand
   ClientUserInfoChanged,     //pfnClientUserInfoChanged
   ServerActivate,            //pfnServerActivate
   ServerDeactivate,          //pfnServerDeactivate

   PlayerPreThink,            //pfnPlayerPreThink
   PlayerPostThink,           //pfnPlayerPostThink

   StartFrame,                //pfnStartFrame
   ParmsNewLevel,             //pfnParmsNewLevel
   ParmsChangeLevel,          //pfnParmsChangeLevel

   GetGameDescription,        //pfnGetGameDescription    Returns string describing current .dll game.
   PlayerCustomization,       //pfnPlayerCustomization   Notifies .dll of new customization for player.

   SpectatorConnect,          //pfnSpectatorConnect      Called when spectator joins server
   SpectatorDisconnect,       //pfnSpectatorDisconnect   Called when spectator leaves the server
   SpectatorThink,            //pfnSpectatorThink        Called when spectator sends a command packet (usercmd_t)

   Sys_Error,                 //pfnSys_Error          Called when engine has encountered an error

   PM_Move,                   //pfnPM_Move
   PM_Init,                   //pfnPM_Init            Server version of player movement initialization
   PM_FindTextureType,        //pfnPM_FindTextureType

   SetupVisibility,           //pfnSetupVisibility        Set up PVS and PAS for networking for this client
   UpdateClientData,          //pfnUpdateClientData       Set up data sent only to specific client
   AddToFullPack,             //pfnAddToFullPack
   CreateBaseline,            //pfnCreateBaseline        Tweak entity baseline for network encoding, allows setup of player baselines, too.
   RegisterEncoders,          //pfnRegisterEncoders      Callbacks for network encoding
   GetWeaponData,             //pfnGetWeaponData
   CmdStart,                  //pfnCmdStart
   CmdEnd,                    //pfnCmdEnd
   ConnectionlessPacket,      //pfnConnectionlessPacket
   GetHullBounds,             //pfnGetHullBounds
   CreateInstancedBaselines,  //pfnCreateInstancedBaselines
   InconsistentFile,          //pfnInconsistentFile
   AllowLagCompensation,      //pfnAllowLagCompensation
};

extern "C" EXPORT int GetEntityAPI( DLL_FUNCTIONS *pFunctionTable, int interfaceVersion )
{
   // check if engine's pointer is valid and version is correct...

   if ( !pFunctionTable || interfaceVersion != INTERFACE_VERSION )
      return FALSE;

   // pass engine callback function table to engine...
   memcpy( pFunctionTable, &gFunctionTable, sizeof( DLL_FUNCTIONS ) );

   // call each bot dll GetEntityAPI function...
   for (int index = 0; index < num_bot_dlls; index++)
   {
      if (!(*bot_dlls[index].other_GetEntityAPI)(&bot_dlls[index].other_gFunctionTable, INTERFACE_VERSION))
      {
         return FALSE;
      }
   }

   // pass other DLLs engine callbacks to function table...
   if (!(*other_GetEntityAPI)(&other_gFunctionTable, INTERFACE_VERSION))
   {
      return FALSE;  // error initializing function table!!!
   }

   return TRUE;
}


BOOL FakeDLLClientConnect( edict_t *pEntity, const char *pszName, const char *pszAddress, char szRejectReason[ 128 ]  )
{ 
   return (*other_gFunctionTable.pfnClientConnect)(pEntity, pszName, pszAddress, szRejectReason);
}

void FakeDLLClientPutInServer( edict_t *pEntity )
{
   (*other_gFunctionTable.pfnClientPutInServer)(pEntity);
}

void FakeDLLUpdateClientData ( const struct edict_s *ent, int sendweapons, struct clientdata_s *cd )
{
   (*other_gFunctionTable.pfnUpdateClientData)(ent, sendweapons, cd);
}

int FakeDLLGetWeaponData( struct edict_s *player, struct weapon_data_s *info )
{
   return (*other_gFunctionTable.pfnGetWeaponData)(player, info);
}

void FakeDLLCmdStart( const edict_t *player, const struct usercmd_s *cmd, unsigned int random_seed )
{
   (*other_gFunctionTable.pfnCmdStart)(player, cmd, random_seed);
}

void FakeDLLCmdEnd ( const edict_t *player )
{
   (*other_gFunctionTable.pfnCmdEnd)(player);
}



DLL_FUNCTIONS gFakeDLLFunctionTable =
{
   NULL,   //pfnGameInit
   NULL,   //pfnSpawn
   NULL,   //pfnThink
   NULL,   //pfnUse
   NULL,   //pfnTouch
   NULL,   //pfnBlocked
   NULL,   //pfnKeyValue
   NULL,   //pfnSave
   NULL,   //pfnRestore
   NULL,   //pfnAbsBox

   NULL,   //pfnSaveWriteFields
   NULL,   //pfnSaveReadFields

   NULL,   //pfnSaveGlobalState
   NULL,   //pfnRestoreGlobalState
   NULL,   //pfnResetGlobalState

   FakeDLLClientConnect,   //pfnClientConnect
   NULL,   //pfnClientDisconnect
   NULL,   //pfnClientKill
   FakeDLLClientPutInServer,   //pfnClientPutInServer
   NULL,   //pfnClientCommand
   NULL,   //pfnClientUserInfoChanged
   NULL,   //pfnServerActivate
   NULL,   //pfnServerDeactivate

   NULL,   //pfnPlayerPreThink
   NULL,   //pfnPlayerPostThink

   NULL,   //pfnStartFrame
   NULL,   //pfnParmsNewLevel
   NULL,   //pfnParmsChangeLevel
   NULL,   //pfnGetGameDescription    Returns string describing current .dll game.
   NULL,   //pfnPlayerCustomization   Notifies .dll of new customization for player.

   NULL,   //pfnSpectatorConnect      Called when spectator joins server
   NULL,   //pfnSpectatorDisconnect   Called when spectator leaves the server
   NULL,   //pfnSpectatorThink        Called when spectator sends a command packet (usercmd_t)

   NULL,   //pfnSys_Error          Called when engine has encountered an error

   NULL,   //pfnPM_Move
   NULL,   //pfnPM_Init            Server version of player movement initialization
   NULL,   //pfnPM_FindTextureType

   NULL,   //pfnSetupVisibility        Set up PVS and PAS for networking for this client
   FakeDLLUpdateClientData,   //pfnUpdateClientData       Set up data sent only to specific client
   NULL,   //pfnAddToFullPack
   NULL,   //pfnCreateBaseline        Tweak entity baseline for network encoding, allows setup of player baselines, too.
   NULL,   //pfnRegisterEncoders      Callbacks for network encoding
   FakeDLLGetWeaponData,   //pfnGetWeaponData
   FakeDLLCmdStart,        //pfnCmdStart
   FakeDLLCmdEnd,          //pfnCmdEnd
   NULL,  //pfnConnectionlessPacket
   NULL,  //pfnGetHullBounds
   NULL,  //pfnCreateInstancedBaselines
   NULL,  //pfnInconsistentFile
   NULL,  //pfnAllowLagCompensation
};


extern "C" EXPORT int GetFakeDLLEntityAPI( DLL_FUNCTIONS *pFunctionTable, int interfaceVersion )
{
   memcpy( pFunctionTable, &gFakeDLLFunctionTable, sizeof( DLL_FUNCTIONS ) );
   
   return TRUE;
}


extern "C" EXPORT int GetNewDLLFunctions( NEW_DLL_FUNCTIONS *pFunctionTable, int *interfaceVersion )
{
   if (other_GetNewDLLFunctions == NULL)
      return FALSE;

   // call each bot dll GetNewDLLFunctions function...
   for (int index = 0; index < num_bot_dlls; index++)
   {
      if (!(*bot_dlls[index].other_GetNewDLLFunctions)(pFunctionTable, interfaceVersion))
      {
         return FALSE;
      }
   }

   // pass other DLLs engine callbacks to function table...
   if (!(*other_GetNewDLLFunctions)(pFunctionTable, interfaceVersion))
   {
      return FALSE;  // error initializing function table!!!
   }

   return TRUE;
}


extern "C" EXPORT void MasterFakeClientCommand(edict_t *pEntity, char *cmd)
{
   int cnt, index, pos;

   if (debug_engine) { fp=fopen("bot_mixer.txt","a"); fprintf(fp,"MasterFakeClientCommand=%s\n",cmd); fclose(fp); }

   cnt = 0;
   index = 0;

   if (cmd == NULL)
      return;

   if (*cmd == 0)
      return;

   memset(g_argv, 0, 1024);

   // skip any leading blanks
   while ((cmd[index] == ' ') && (cmd[index]))
      index++;

   pos = 0;

   while (cmd[index])
   {
      // process non-blank characters
      while ((cmd[index] != ' ') && (cmd[index]))
      {
         g_argv[512+pos] = cmd[index];
         g_argv[pos++] = cmd[index++];
      }

      g_argv[512+pos] = 0;  // null terminate argv arguments
      g_argv[pos] = 0;  // null terminate argv string

      cnt++;  // count argument separators

      // skip blanks
      while ((cmd[index] == ' ') && (cmd[index]))
         index++;

      if (cmd[index])
         g_argv[pos] = ' ';  // add blank separator between arguments

      pos++;
   }

   g_argv[pos] = 0;  // null terminate argv string

   fake_arg_count = cnt;

   isFakeClientCommand = 1;

   // allow the MOD DLL to execute the ClientCommand...
   (*other_gFunctionTable.pfnClientCommand)(pEntity);

   isFakeClientCommand = 0;
}


void BotFakeClientCommand(int bot_dll_index, edict_t *pEntity, char *cmd)
{
   int cnt, index, pos;

   if (debug_engine) { fp=fopen("bot_mixer.txt","a"); fprintf(fp,"BotFakeClientCommand=%s\n",cmd); fclose(fp); }

   cnt = 0;
   index = 0;

   if (cmd == NULL)
      return;

   if (*cmd == 0)
      return;

   memset(g_argv, 0, 1024);

   // skip any leading blanks
   while ((cmd[index] == ' ') && (cmd[index]))
      index++;

   pos = 0;

   while (cmd[index])
   {
      // process non-blank characters
      while ((cmd[index] != ' ') && (cmd[index]))
      {
         g_argv[512+pos] = cmd[index];
         g_argv[pos++] = cmd[index++];
      }

      g_argv[512+pos] = 0;  // null terminate argv arguments
      g_argv[pos] = 0;  // null terminate argv string

      cnt++;  // count argument separators

      // skip blanks
      while ((cmd[index] == ' ') && (cmd[index]))
         index++;

      if (cmd[index])
         g_argv[pos] = ' ';  // add blank separator between arguments

      pos++;
   }

   g_argv[pos] = 0;  // null terminate argv string

   fake_arg_count = cnt;

   isFakeClientCommand = 1;

   int dll_index = current_dll_index;  // save current dll index

   FakeClientSet(bot_dll_index);

   (*bot_dlls[bot_dll_index].other_gFunctionTable.pfnClientCommand)(pEntity);

   current_dll_index = dll_index;  // restore
   FakeClientRestore();

   isFakeClientCommand = 0;
}


const char *Cmd_Args( void )
{
   if (isFakeClientCommand)
   {
      return &g_argv[0];
   }
   else
   {
      return (*g_engfuncs.pfnCmd_Args)();
   }
}


const char *Cmd_Argv( int argc )
{
   int cnt, index;

   if (isFakeClientCommand)
   {
      cnt = argc;
      index = 512;  // start at null terminated arguments

      while (g_argv[index])
      {
         if (cnt == 0)
            return &g_argv[index];

         // skip non-null characters
         while (g_argv[index] != 0)
            index++;

         // skip null character
         index++;

         cnt--;  // decrement argument counter
      }

      return NULL;
   }
   else
   {
      return (*g_engfuncs.pfnCmd_Argv)(argc);
   }
}


int Cmd_Argc( void )
{
   if (isFakeClientCommand)
   {
      return fake_arg_count;
   }
   else
   {
      return (*g_engfuncs.pfnCmd_Argc)();
   }
}


extern "C" void player( entvars_t *pev );

extern "C" EXPORT void master_player( entvars_t *pev ) 
{
   player(pev);
}
