//
// BotMixer5000 - botman's multibot loader
//
// (http://planethalflife.com/botman/)
//
// h_export.cpp
//

#include "extdll.h"
#include "enginecallback.h"
#include "util.h"
#include "cbase.h"
#include "engine.h"

#include "bot_mixer.h"

#ifndef __linux__
HGLOBAL h_global_argv = NULL;
#else
char h_global_argv[1024];
#endif

BotLibrary h_Library = NULL;

enginefuncs_t g_engfuncs;
globalvars_t  *gpGlobals;
enginefuncs_t g_bot_engfuncs;
char *g_argv;

bot_dlls_t bot_dlls[MAX_DLLS];
int num_bot_dlls;

static FILE *fp;


GETENTITYAPI other_GetEntityAPI = NULL;
GETNEWDLLFUNCTIONS other_GetNewDLLFunctions = NULL;
GIVEFNPTRSTODLL other_GiveFnptrsToDll = NULL;


#ifndef __linux__

// Required DLL entry point
BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
   if (fdwReason == DLL_PROCESS_ATTACH)
   {
   }
   else if (fdwReason == DLL_PROCESS_DETACH)
   {
      if (h_Library)
         FreeLibrary(h_Library);

      if (h_global_argv)
      {
         GlobalUnlock(h_global_argv);
         GlobalFree(h_global_argv);
      }
   }

   return TRUE;
}

#endif

#ifndef __linux__
void DLLEXPORT GiveFnptrsToDll( enginefuncs_t* pengfuncsFromEngine, globalvars_t *pGlobals )
#else
extern "C" DLLEXPORT GiveFnptrsToDll( enginefuncs_t* pengfuncsFromEngine, globalvars_t *pGlobals )
#endif
{
   int pos;
   char game_dir[256];
   char mod_name[32];
   FILE *fpConfig;
   char message[256];
   char filename[256];
   char mod_dllname[64];
   char mod_dllfile[256];
   char input[256];
   int index, len;


   // get the engine functions from the engine...

   memcpy(&g_engfuncs, pengfuncsFromEngine, sizeof(enginefuncs_t));
   gpGlobals = pGlobals;

   // find the directory name of the currently running MOD...
   (*g_engfuncs.pfnGetGameDir)(game_dir);

   pos = strlen(game_dir) - 1;

   // scan backwards till first directory separator...
   while ((pos) && (game_dir[pos] != '/'))
      pos--;

   if (pos == 0)
   {
      // Error getting directory name!

      ALERT( at_error, "BotMixer5000 - Error determining MOD directory name!" );
      printf("BotMixer5000 - Error determining MOD directory name!\n");
      Sleep(5000);
      exit(0);
   }

   pos++;
   strcpy(mod_name, &game_dir[pos]);

   for (index = 0; index < MAX_DLLS; index++)
   {
      memset(bot_dlls[index].cmd, 0, sizeof(bot_dlls[index].cmd));
      memset(bot_dlls[index].dll_filename, 0, sizeof(bot_dlls[index].dll_filename));
      bot_dlls[index].other_GetEntityAPI = NULL;
      bot_dlls[index].other_GetNewDLLFunctions = NULL;
      bot_dlls[index].other_GiveFnptrsToDll = NULL;
   }

   // open the botmixer5000.cfg file...

   strcpy(filename, mod_name);
   strcat(filename, "/");
   strcat(filename, "botmixer5000.cfg");

   UTIL_Pathname_Convert(filename);

   fpConfig = fopen(filename, "r");

   if (fpConfig == NULL)
   {
      // Error opening config file

      ALERT( at_error, "BotMixer5000 - Can't open botmixer5000.cfg file!" );
      printf("BotMixer5000 - Can't open botmixer5000.cfg file!");
      Sleep(5000);
      exit(0);
   }

   fgets(mod_dllname, sizeof(mod_dllname), fpConfig);

   UTIL_TrimWhitespace(mod_dllname);  // trim leading and trailing whitespace

   // now read the command line alias and the bot DLL filename...

   num_bot_dlls = 0;
   while (!feof(fpConfig))
   {
      if (fgets(input, sizeof(input), fpConfig) != NULL)
      {
         UTIL_TrimWhitespace(input);  // trim leading and trailing whitespace
         UTIL_ReplaceWhitespace(input);  // remove duplicate whitespace

         len = strlen(input);

         if (len > 1)
         {
            // ignore comment lines
            if ((input[0] == '/') && (input[1] == '/'))
               continue;
         }

         if ((input[0]) && (num_bot_dlls < MAX_DLLS))
         {
            char *pos = strstr(input, " ");

            if (pos)
            {
               *pos = 0;

               strcpy(bot_dlls[num_bot_dlls].cmd, input);
               strcpy(bot_dlls[num_bot_dlls].dll_filename, pos+1);

               num_bot_dlls++;
            }
         }
      }
   }


   fclose(fpConfig);

   strcpy(mod_dllfile, mod_name);
   strcat(mod_dllfile, "/");
   strcat(mod_dllfile, mod_dllname);  // includes "dlls/" before dll filename

   UTIL_Pathname_Convert(mod_dllfile);

   sprintf(message, "BotMixer5000: loading %s\n", mod_dllfile);

   ALERT( at_console, message);

#ifndef __linux__
   h_Library = LoadLibrary(mod_dllfile);
#else
   h_Library = dlopen(mod_dllfile, RTLD_NOW);
#endif

   if (h_Library == NULL)
   {
      // Directory error or Unsupported MOD!

      sprintf(message, "BotMixer5000 - MOD dll (%s) not found (or unsupported MOD)!",mod_dllfile);
      ALERT( at_error, message);
      printf("%s\n", message);
      Sleep(5000);
      exit(0);
   }


#ifndef __linux__
   h_global_argv = GlobalAlloc(GMEM_SHARE, 1024);
   g_argv = (char *)GlobalLock(h_global_argv);
#else
   g_argv = (char *)h_global_argv;
#endif


   other_GetEntityAPI = (GETENTITYAPI)GetProcAddress(h_Library, "GetEntityAPI");

   if (other_GetEntityAPI == NULL)
   {
      // Can't find GetEntityAPI!

      ALERT( at_error, "BotMixer5000 - Can't get MOD's GetEntityAPI!" );
   }

   other_GetNewDLLFunctions = (GETNEWDLLFUNCTIONS)GetProcAddress(h_Library, "GetNewDLLFunctions");

//   if (other_GetNewDLLFunctions == NULL)
//   {
//      // Can't find GetNewDLLFunctions!
//   }

   other_GiveFnptrsToDll = (GIVEFNPTRSTODLL)GetProcAddress(h_Library, "GiveFnptrsToDll");

   if (other_GiveFnptrsToDll == NULL)
   {
      // Can't find GiveFnptrsToDll!

      ALERT( at_error, "BotMixer5000 - Can't get MOD's GiveFnptrsToDll!" );
   }

   for (index = 0; index < num_bot_dlls; index++)
   {
      strcpy(mod_dllfile, mod_name);
      strcat(mod_dllfile, "/");
      strcat(mod_dllfile, bot_dlls[index].dll_filename);

      UTIL_Pathname_Convert(mod_dllfile);

      sprintf(message, "BotMixer5000: loading %s\n", mod_dllfile);

      ALERT( at_console, message);

#ifndef __linux__
      bot_dlls[index].h_Library = LoadLibrary(mod_dllfile);
#else
      bot_dlls[index].h_Library = dlopen(mod_dllfile, RTLD_NOW);
#endif

      if (bot_dlls[index].h_Library == NULL)
      {
         sprintf(message, "BotMixer5000 - bot dll (%s) not found!", mod_dllfile);
         ALERT( at_error, message);
         printf("%s\n", message);
         Sleep(5000);
         exit(0);
      }

      bot_dlls[index].other_GetEntityAPI = 
         (GETENTITYAPI)GetProcAddress(bot_dlls[index].h_Library, "GetEntityAPI");

      bot_dlls[index].other_GetNewDLLFunctions =
         (GETNEWDLLFUNCTIONS)GetProcAddress(bot_dlls[index].h_Library, "GetNewDLLFunctions");

      bot_dlls[index].other_GiveFnptrsToDll =
         (GIVEFNPTRSTODLL)GetProcAddress(bot_dlls[index].h_Library, "GiveFnptrsToDll");
   }

   pengfuncsFromEngine->pfnCmd_Args = Cmd_Args;
   pengfuncsFromEngine->pfnCmd_Argv = Cmd_Argv;
   pengfuncsFromEngine->pfnCmd_Argc = Cmd_Argc;

   pengfuncsFromEngine->pfnPrecacheModel = pfnPrecacheModel;
   pengfuncsFromEngine->pfnPrecacheSound = pfnPrecacheSound;
   pengfuncsFromEngine->pfnSetModel = pfnSetModel;
   pengfuncsFromEngine->pfnModelIndex = pfnModelIndex;
   pengfuncsFromEngine->pfnModelFrames = pfnModelFrames;
   pengfuncsFromEngine->pfnSetSize = pfnSetSize;
   pengfuncsFromEngine->pfnChangeLevel = pfnChangeLevel;
   pengfuncsFromEngine->pfnGetSpawnParms = pfnGetSpawnParms;
   pengfuncsFromEngine->pfnSaveSpawnParms = pfnSaveSpawnParms;
   pengfuncsFromEngine->pfnVecToYaw = pfnVecToYaw;
   pengfuncsFromEngine->pfnVecToAngles = pfnVecToAngles;
   pengfuncsFromEngine->pfnMoveToOrigin = pfnMoveToOrigin;
   pengfuncsFromEngine->pfnChangeYaw = pfnChangeYaw;
   pengfuncsFromEngine->pfnChangePitch = pfnChangePitch;
   pengfuncsFromEngine->pfnFindEntityByString = pfnFindEntityByString;
   pengfuncsFromEngine->pfnGetEntityIllum = pfnGetEntityIllum;
   pengfuncsFromEngine->pfnFindEntityInSphere = pfnFindEntityInSphere;
   pengfuncsFromEngine->pfnFindClientInPVS = pfnFindClientInPVS;
   pengfuncsFromEngine->pfnEntitiesInPVS = pfnEntitiesInPVS;
   pengfuncsFromEngine->pfnMakeVectors = pfnMakeVectors;
   pengfuncsFromEngine->pfnAngleVectors = pfnAngleVectors;
   pengfuncsFromEngine->pfnCreateEntity = pfnCreateEntity;
   pengfuncsFromEngine->pfnRemoveEntity = pfnRemoveEntity;
   pengfuncsFromEngine->pfnCreateNamedEntity = pfnCreateNamedEntity;
   pengfuncsFromEngine->pfnMakeStatic = pfnMakeStatic;
   pengfuncsFromEngine->pfnEntIsOnFloor = pfnEntIsOnFloor;
   pengfuncsFromEngine->pfnDropToFloor = pfnDropToFloor;
   pengfuncsFromEngine->pfnWalkMove = pfnWalkMove;
   pengfuncsFromEngine->pfnSetOrigin = pfnSetOrigin;
   pengfuncsFromEngine->pfnEmitSound = pfnEmitSound;
   pengfuncsFromEngine->pfnEmitAmbientSound = pfnEmitAmbientSound;
   pengfuncsFromEngine->pfnTraceLine = pfnTraceLine;
   pengfuncsFromEngine->pfnTraceToss = pfnTraceToss;
   pengfuncsFromEngine->pfnTraceMonsterHull = pfnTraceMonsterHull;
   pengfuncsFromEngine->pfnTraceHull = pfnTraceHull;
   pengfuncsFromEngine->pfnTraceModel = pfnTraceModel;
   pengfuncsFromEngine->pfnTraceTexture = pfnTraceTexture;
   pengfuncsFromEngine->pfnTraceSphere = pfnTraceSphere;
   pengfuncsFromEngine->pfnGetAimVector = pfnGetAimVector;
   pengfuncsFromEngine->pfnServerCommand = pfnServerCommand;
   pengfuncsFromEngine->pfnServerExecute = pfnServerExecute;

   // don't override the var args stuff!!!
//   pengfuncsFromEngine->pfnClientCommand = pfnClientCommand;

   pengfuncsFromEngine->pfnParticleEffect = pfnParticleEffect;
   pengfuncsFromEngine->pfnLightStyle = pfnLightStyle;
   pengfuncsFromEngine->pfnDecalIndex = pfnDecalIndex;
   pengfuncsFromEngine->pfnPointContents = pfnPointContents;
   pengfuncsFromEngine->pfnMessageBegin = pfnMessageBegin;
   pengfuncsFromEngine->pfnMessageEnd = pfnMessageEnd;
   pengfuncsFromEngine->pfnWriteByte = pfnWriteByte;
   pengfuncsFromEngine->pfnWriteChar = pfnWriteChar;
   pengfuncsFromEngine->pfnWriteShort = pfnWriteShort;
   pengfuncsFromEngine->pfnWriteLong = pfnWriteLong;
   pengfuncsFromEngine->pfnWriteAngle = pfnWriteAngle;
   pengfuncsFromEngine->pfnWriteCoord = pfnWriteCoord;
   pengfuncsFromEngine->pfnWriteString = pfnWriteString;
   pengfuncsFromEngine->pfnWriteEntity = pfnWriteEntity;
   pengfuncsFromEngine->pfnCVarRegister = pfnCVarRegister;
   pengfuncsFromEngine->pfnCVarGetFloat = pfnCVarGetFloat;
   pengfuncsFromEngine->pfnCVarGetString = pfnCVarGetString;
   pengfuncsFromEngine->pfnCVarSetFloat = pfnCVarSetFloat;
   pengfuncsFromEngine->pfnCVarSetString = pfnCVarSetString;
   pengfuncsFromEngine->pfnPvAllocEntPrivateData = pfnPvAllocEntPrivateData;
   pengfuncsFromEngine->pfnPvEntPrivateData = pfnPvEntPrivateData;
   pengfuncsFromEngine->pfnFreeEntPrivateData = pfnFreeEntPrivateData;
   pengfuncsFromEngine->pfnSzFromIndex = pfnSzFromIndex;
   pengfuncsFromEngine->pfnAllocString = pfnAllocString;
   pengfuncsFromEngine->pfnGetVarsOfEnt = pfnGetVarsOfEnt;
   pengfuncsFromEngine->pfnPEntityOfEntOffset = pfnPEntityOfEntOffset;
   pengfuncsFromEngine->pfnEntOffsetOfPEntity = pfnEntOffsetOfPEntity;
   pengfuncsFromEngine->pfnIndexOfEdict = pfnIndexOfEdict;
   pengfuncsFromEngine->pfnPEntityOfEntIndex = pfnPEntityOfEntIndex;
   pengfuncsFromEngine->pfnFindEntityByVars = pfnFindEntityByVars;
   pengfuncsFromEngine->pfnGetModelPtr = pfnGetModelPtr;
   pengfuncsFromEngine->pfnRegUserMsg = pfnRegUserMsg;
   pengfuncsFromEngine->pfnAnimationAutomove = pfnAnimationAutomove;
   pengfuncsFromEngine->pfnGetBonePosition = pfnGetBonePosition;
   pengfuncsFromEngine->pfnFunctionFromName = pfnFunctionFromName;
   pengfuncsFromEngine->pfnNameForFunction = pfnNameForFunction;
   pengfuncsFromEngine->pfnClientPrintf = pfnClientPrintf;
   pengfuncsFromEngine->pfnServerPrint = pfnServerPrint;
   pengfuncsFromEngine->pfnGetAttachment = pfnGetAttachment;
   pengfuncsFromEngine->pfnCRC32_Init = pfnCRC32_Init;
   pengfuncsFromEngine->pfnCRC32_ProcessBuffer = pfnCRC32_ProcessBuffer;
   pengfuncsFromEngine->pfnCRC32_ProcessByte = pfnCRC32_ProcessByte;
   pengfuncsFromEngine->pfnCRC32_Final = pfnCRC32_Final;
   pengfuncsFromEngine->pfnRandomLong = pfnRandomLong;
   pengfuncsFromEngine->pfnRandomFloat = pfnRandomFloat;
   pengfuncsFromEngine->pfnSetView = pfnSetView;
   pengfuncsFromEngine->pfnTime = pfnTime;
   pengfuncsFromEngine->pfnCrosshairAngle = pfnCrosshairAngle;
   pengfuncsFromEngine->pfnLoadFileForMe = pfnLoadFileForMe;
   pengfuncsFromEngine->pfnFreeFile = pfnFreeFile;
   pengfuncsFromEngine->pfnEndSection = pfnEndSection;
   pengfuncsFromEngine->pfnCompareFileTime = pfnCompareFileTime;
   pengfuncsFromEngine->pfnGetGameDir = pfnGetGameDir;
   pengfuncsFromEngine->pfnCvar_RegisterVariable = pfnCvar_RegisterVariable;
   pengfuncsFromEngine->pfnFadeClientVolume = pfnFadeClientVolume;
   pengfuncsFromEngine->pfnSetClientMaxspeed = pfnSetClientMaxspeed;
   pengfuncsFromEngine->pfnCreateFakeClient = pfnCreateFakeClient;
   pengfuncsFromEngine->pfnRunPlayerMove = pfnRunPlayerMove;
   pengfuncsFromEngine->pfnNumberOfEntities = pfnNumberOfEntities;
   pengfuncsFromEngine->pfnGetInfoKeyBuffer = pfnGetInfoKeyBuffer;
   pengfuncsFromEngine->pfnInfoKeyValue = pfnInfoKeyValue;
   pengfuncsFromEngine->pfnSetKeyValue = pfnSetKeyValue;
   pengfuncsFromEngine->pfnSetClientKeyValue = pfnSetClientKeyValue;
   pengfuncsFromEngine->pfnIsMapValid = pfnIsMapValid;
   pengfuncsFromEngine->pfnStaticDecal = pfnStaticDecal;
   pengfuncsFromEngine->pfnPrecacheGeneric = pfnPrecacheGeneric;
   pengfuncsFromEngine->pfnGetPlayerUserId = pfnGetPlayerUserId;
   pengfuncsFromEngine->pfnBuildSoundMsg = pfnBuildSoundMsg;
   pengfuncsFromEngine->pfnIsDedicatedServer = pfnIsDedicatedServer;
   pengfuncsFromEngine->pfnCVarGetPointer = pfnCVarGetPointer;
   pengfuncsFromEngine->pfnGetPlayerWONId = pfnGetPlayerWONId;

   memcpy(&g_bot_engfuncs, pengfuncsFromEngine, sizeof(enginefuncs_t));

   for (index = 0; index < num_bot_dlls; index++)
   {
      // here we do some REALLY nasty, kludgey, sneaky stuff to chain
      // the engine function callbacks from one bot to the next...
      //
      // call bot GiveFnptrsToDLL which will in turn call fakedll.dll
      // version of GiveFnptrsToDLL which will call GiveBotFnptrsToMaster
      // in this DLL where we will copy the bot's version of engine
      // functions back into g_bot_engfuncs array.  This way we can
      // chain the engine functions so that it will look like this...
      //
      // engine <- botmixer5000 <- bot1 <- bot2 <- bot3 <- MOD
      // 
      // (where "<-" indicates the direction of the function calls.)
      //
      // the return value from the engine function will get propagated
      // back up in the reverse direction...
      //
      // engine -> botmixer5000 -> bot1 -> bot2 -> bot3 -> MOD
      //

      (*bot_dlls[index].other_GiveFnptrsToDll)(&g_bot_engfuncs, pGlobals);

      // note that GiveFnptrsToMaster should get called from the previous
      // function so the contents of g_bot_engfuncs will actually be
      // changing inside this for loop even though this function it not
      // the one modifying the contents.
   }

   // give the engine functions to the MOD DLL...
   (*other_GiveFnptrsToDll)(&g_bot_engfuncs, pGlobals);
}


#ifndef __linux__
void DLLEXPORT GiveBotFnptrsToMaster( enginefuncs_t* pengfuncsFromEngine, globalvars_t *pGlobals )
#else
extern "C" DLLEXPORT GiveBotFnptrsToMaster( enginefuncs_t* pengfuncsFromEngine, globalvars_t *pGlobals )
#endif
{
   memcpy(&g_bot_engfuncs, pengfuncsFromEngine, sizeof(enginefuncs_t));
}
