//
// fakedll (for BotMixer5000 - botman's multibot loader)
//
// (http://planethalflife.com/botman/)
//
// h_export.cpp
//

#include "extdll.h"
#include "enginecallback.h"
#include "util.h"
#include "cbase.h"
#include "engine.h"

#include "fakedll.h"


#ifndef __linux__
HGLOBAL h_global_argv = NULL;
#else
char h_global_argv[1024];
#endif


BotLibrary h_Library = NULL;

enginefuncs_t g_engfuncs;
GIVEFNPTRSTODLL master_GiveBotFnptrsToMaster = NULL;
char *g_argv;


void UTIL_Pathname_Convert(char *pathname);


#ifndef __linux__

// Required DLL entry point
BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
   if (fdwReason == DLL_PROCESS_ATTACH)
   {
   }
   else if (fdwReason == DLL_PROCESS_DETACH)
   {
      if (h_Library)
         FreeLibrary(h_Library);

      if (h_global_argv)
      {
         GlobalUnlock(h_global_argv);
         GlobalFree(h_global_argv);
      }
   }

   return TRUE;
}

#endif

#ifndef __linux__
void DLLEXPORT GiveFnptrsToDll( enginefuncs_t* pengfuncsFromEngine, globalvars_t *pGlobals )
#else
extern "C" DLLEXPORT GiveFnptrsToDll( enginefuncs_t* pengfuncsFromEngine, globalvars_t *pGlobals )
#endif
{
   int pos;
   char game_dir[256];
   char mod_name[32];
   char mixer_dllfile[256];

   // get the engine functions from the engine...

   memcpy(&g_engfuncs, pengfuncsFromEngine, sizeof(enginefuncs_t));

   // find the directory name of the currently running MOD...
   (*g_engfuncs.pfnGetGameDir)(game_dir);

   pos = strlen(game_dir) - 1;

   // scan backwards till first directory separator...
   while ((pos) && (game_dir[pos] != '/'))
      pos--;

   if (pos == 0)
   {
      // Error getting directory name!

      ALERT( at_error, "fakedll - Error determining MOD directory name!" );
      printf("fakedll - Error determining MOD directory name!\n");
      Sleep(5000);
      exit(0);
   }

   pos++;
   strcpy(mod_name, &game_dir[pos]);

   strcpy(mixer_dllfile, mod_name);
#ifndef __linux__
   strcat(mixer_dllfile, "/dlls/BotMixer5000.dll");
#else
   strcat(mixer_dllfile, "/dlls/BotMixer5000_i486.so");
#endif

   UTIL_Pathname_Convert(mixer_dllfile);

#ifndef __linux__
   h_Library = LoadLibrary(mixer_dllfile);
#else
   h_Library = dlopen(mixer_dllfile, RTLD_NOW);
#endif

   if (h_Library == NULL)
   {
      // Directory error or Unsupported MOD!

      ALERT( at_error, "fakedll - BotMixer5000 dll not found!");
      printf("fakedll - BotMixer5000 dll not found!\n");
      Sleep(5000);
      exit(0);
   }

#ifndef __linux__
   h_global_argv = GlobalAlloc(GMEM_SHARE, 1024);
   g_argv = (char *)GlobalLock(h_global_argv);
#else
   g_argv = (char *)h_global_argv;
#endif

   master_GiveBotFnptrsToMaster = (GIVEFNPTRSTODLL)GetProcAddress(h_Library, "GiveBotFnptrsToMaster");

   if (master_GiveBotFnptrsToMaster == NULL)
   {
      // Can't find GiveBotFnptrsToMaster!

      ALERT( at_error, "fakedll - Can't get BotMixer's GiveBotFnptrsToMaster!" );
      printf("fakedll - Can't get BotMixer's GiveBotFnptrsToMaster!\n" );
      Sleep(5000);
      exit(0);
   }

   (*master_GiveBotFnptrsToMaster)(pengfuncsFromEngine, pGlobals);

   return;
}


void UTIL_Pathname_Convert(char *pathname)
{
   int index;
   int len = strlen(pathname);

   for (index=0; index < len; index++)
   {
#ifndef __linux__
      if (pathname[index] == '/')  // convert to MS-DOS format
         pathname[index] = '\\';
#else
      if (pathname[index] == '\\')  // convert to Linux format
         pathname[index] = '/';
#endif
   }
}
